use crate::{decks::new_blackjack_shoe, hand::Hand};

pub struct BlackjackGame {
    shoe: Hand,
    players: Vec<Player>,
    dealer_hand: Hand,
    state: GameState,
}

pub struct Player {
    hands: Vec<PlayingHand>,
}

impl Player {
    fn new() -> Self {
        Player { hands: Vec::new() }
    }

    pub fn get_hands(&self) -> &Vec<PlayingHand> {
        &self.hands
    }

    fn next_playing_hand(&self) -> Option<usize> {
        self.hands
            .iter()
            .position(|e| e.state == HandState::Playing)
    }
}

pub struct PlayingHand {
    hand: Hand,
    state: HandState,
}

impl PlayingHand {
    fn new() -> Self {
        PlayingHand {
            hand: Hand::new(),
            state: HandState::Playing,
        }
    }

    pub fn get_hand(&self) -> &Hand {
        &self.hand
    }

    pub fn get_state(&self) -> HandState {
        self.state
    }
}

#[derive(Clone, Copy, Debug, PartialEq)]
pub enum HandState {
    Playing,
    Standing,
    DoubleDown,
    Busted,
    Blackjack,
    Maxed, // Reached 21
}

#[derive(Clone, Copy, Debug)]
pub enum PlayMoves {
    Hit,
    Stand,
    DoubleDown,
    Deal(usize),
}

pub enum GameState {
    Starting,
    Over,                     // Game is over
    PlayerTurn(usize, usize), // Its the turn of the player
}

impl BlackjackGame {
    pub fn new() -> Self {
        BlackjackGame {
            shoe: new_blackjack_shoe(6),
            players: Vec::new(),
            dealer_hand: Hand::new(),
            state: GameState::Starting,
        }
    }

    pub fn get_dealer_hand(&self) -> &Hand {
        &self.dealer_hand
    }

    pub fn get_state(&self) -> &GameState {
        &self.state
    }

    pub fn get_player(&self, index: usize) -> Option<&Player> {
        self.players.get(index)
    }

    pub fn get_player_count(&self) -> usize {
        self.players.len()
    }

    pub fn play(&mut self, action: PlayMoves) -> bool {
        match (&self.state, action) {
            (GameState::PlayerTurn(player_index, hand_index), PlayMoves::Hit) => {
                let Some(player) = self.players.get_mut(*player_index) else {
                    // Player does not exists
                    return false;
                };

                let Some(hand) = player.hands.get_mut(*hand_index) else {
                    // Hand does not exist
                    return false;
                };

                if hand.state != HandState::Playing {
                    return false;
                }

                hand.hand.add_card(self.shoe.pop_card().unwrap());

                match hand.hand.get_blackjack_value().cmp(&21) {
                    std::cmp::Ordering::Equal => {
                        hand.state = HandState::Maxed;
                    }
                    std::cmp::Ordering::Greater => {
                        hand.state = HandState::Busted;
                    }
                    std::cmp::Ordering::Less => {
                        // Player is still playing
                    }
                };
            }
            (GameState::PlayerTurn(player_index, hand_index), PlayMoves::DoubleDown) => {
                let Some(player) = self.players.get_mut(*player_index) else {
                    // Player does not exists
                    return false;
                };

                let Some(hand) = player.hands.get_mut(*hand_index) else {
                    // Hand does not exist
                    return false;
                };

                if hand.state != HandState::Playing {
                    return false;
                }

                if hand.hand.count() != 2 {
                    // Can oly double down as your first move
                    return false;
                }

                hand.hand.add_card(self.shoe.pop_card().unwrap());

                match hand.hand.get_blackjack_value().cmp(&21) {
                    std::cmp::Ordering::Equal => {
                        hand.state = HandState::Maxed;
                    }
                    std::cmp::Ordering::Less => {
                        hand.state = HandState::Standing;
                    }
                    std::cmp::Ordering::Greater => {
                        hand.state = HandState::Busted;
                    }
                };
            }
            (GameState::PlayerTurn(player_index, hand_index), PlayMoves::Stand) => {
                let Some(player) = self.players.get_mut(*player_index) else {
                    // Player does not exists
                    return false;
                };

                let Some(hand) = player.hands.get_mut(*hand_index) else {
                    // Hand does not exist
                    return false;
                };

                if hand.state != HandState::Playing {
                    // Can only stand on playing hands
                    return false;
                }

                hand.state = HandState::Standing;
            }
            (GameState::Over, PlayMoves::Deal(player_count))
            | (GameState::Starting, PlayMoves::Deal(player_count)) => {

                // Create players
                for _ in 0..player_count{
                    self.players.push(Player::new());
                }

                // Create one hand for the players
                for player in self.players.iter_mut(){
                    player.hands.push(PlayingHand::new());
                }

                // Add one card to each hand the players have
                for player in self.players.iter_mut(){
                    for hand in player.hands.iter_mut(){
                        hand.hand.add_card(self.shoe.pop_card().unwrap());
                    }
                }

                // Add one card to the dealer
                self.dealer_hand.add_card(self.shoe.pop_card().unwrap());

                // Add the 2nd card to the players hand
                for player in self.players.iter_mut(){
                    for hand in player.hands.iter_mut(){
                        hand.hand.add_card(self.shoe.pop_card().unwrap());

                        if hand.hand.is_backjack() {
                            hand.state = HandState::Blackjack;
                        }
                    }
                }

                // Add 2nd card to the dealer
                self.dealer_hand.add_card(self.shoe.pop_card().unwrap());

                self.state = GameState::PlayerTurn(0, 0);
            }
            (_, PlayMoves::Deal(_)) | (GameState::Over, _) | (GameState::Starting, _) => {
                return false;
            }
        }

        // Find next player or dealer turn
        if let Some(next_turn) = self.next_player_and_hand() {
            self.state = GameState::PlayerTurn(next_turn.0, next_turn.1);
        } else {
            self.dealer_turn();
        }

        return true;
    }

    fn dealer_turn(&mut self) {
        loop {
            let dealer_value = self.dealer_hand.get_blackjack_value();
            if dealer_value >= 17 {
                // Dealer stands
                self.state = GameState::Over;
                return;
            }

            self.dealer_hand.add_card(self.shoe.pop_card().unwrap());

            if self.dealer_hand.get_blackjack_value() > 21 {
                // Dealer busts
                self.state = GameState::Over;
                return;
            }
        }
    }

    fn next_player_and_hand(&self) -> Option<(usize, usize)> {
        if let GameState::PlayerTurn(player_index, hand_index) = self.state {
            let Some(player) = self.players.get(player_index) else {
                // Player does not exists
                return None;
            };

            let Some(hand) = player.hands.get(hand_index) else {
                // Hand does not exist
                return None;
            };

            if hand.state == HandState::Playing {
                // Hand is still playing
                return Some((player_index, hand_index));
            }

            // Get next valid hand for player
            if let Some(next_hand) = player.next_playing_hand() {
                return Some((player_index, next_hand));
            }

            // Get next valid player
            if let Some(next_hand) = self.players.iter().enumerate().find_map(|p| {
                if let Some(h) = p.1.next_playing_hand(){
                    Some((p.0,h))
                }else{
                    None
                }
            }) {
                return Some((next_hand.0, next_hand.1));
            }
        }

        None
    }
}
